# -*- coding: utf-8 -*-

class TranslationBase:
    def __init__(self):
        pass
    
    
    # Override this method if you want to modify the given layer,
    # or return None if you want to suppress the layer
    def filter_layer(self, layer):
        return layer
    
    
    # Override this method if you want to modify the given feature,
    # or return None if you want to suppress the feature
    # note 1: fieldNames parameter has been removed from the original ogr2osm,
    # but can be recovered from the ogrfeature parameter:
    # ---
    # feature_def = ogrfeature.GetDefnRef()
    # field_names = []
    # for i in range(feature_def.GetFieldCount()):
    #     field_names.append(feature_def.GetFieldDefn(i).GetNameRef())
    # ---
    # note 2: reproject is a function to convert the feature to 4326 projection
    # with coordinates in traditional gis order. However, do not return the
    # reprojected feature since it will be done again in ogr2pbf.
    def filter_feature(self, ogrfeature, reproject):
        return ogrfeature
    
    
    # Override this method if you want to modify or add tags to the xml output
    def filter_tags(self, tags):
        return tags
    
    
    # This method is called after the creation of an OsmGeometry object. The
    # ogr feature and ogr geometry used to create the object are passed as
    # well. Note that any return values will be discarded by ogr2pbf.
    def process_feature_post(self, osmgeometry, ogrfeature, ogrgeometry):
        pass
    
    
    # Override this method if you want to modify the list of nodes, ways or
    # relations, or take any additional actions right before writing the
    # objects to the OSM file. Note that any return values will be discarded
    # by ogr2pbf.
    def process_output(self, osmnodes, osmways, osmrelations):
        pass




# -*- coding: utf-8 -*-

import logging
from lxml import etree

class OsmId:
    element_id_counter = 0
    element_id_counter_incr = -1

    @staticmethod
    def set_id(start_id, is_positive = False):
        OsmId.element_id_counter = start_id
        if is_positive:
            OsmId.element_id_counter_incr = 1


    @staticmethod
    def load_id(filename):
        with open(filename, 'r') as ff:
            OsmId.element_id_counter = int(ff.readline(20))
        logging.info("Starting counter value '%d' read from file '%s'." % \
                     (OsmId.element_id_counter, filename))


    @staticmethod
    def save_id(filename):
        with open(filename, 'w') as ff:
            ff.write(str(OsmId.element_id_counter))
        logging.info("Wrote elementIdCounter '%d' to file '%s'" % \
                     (OsmId.element_id_counter, filename))



class OsmGeometry:
    def __init__(self):
        self.id = self.__get_new_id()
        self.tags = { }
        self.__parents = set()
    
    
    def __get_new_id(self):
        OsmId.element_id_counter += OsmId.element_id_counter_incr
        return OsmId.element_id_counter
    
    
    def get_parents(self):
        return self.__parents
    
    
    def add_tags(self, tags):
        self.tags.update(tags)
    
    
    def replacejwithi(self, i, j):
        pass
    
    
    def to_xml(self, attributes = { }):
        pass
    
    
    def addparent(self, parent):
        self.__parents.add(parent)


    def removeparent(self, parent):
        self.__parents.discard(parent)



class OsmPoint(OsmGeometry):
    def __init__(self, x, y):
        super().__init__()
        self.x = x
        self.y = y
    
    
    def to_xml(self, attributes):
        xmlattrs = { 'visible':'true', \
                     'id':str(self.id), \
                     'lat':str(self.y),
                     'lon':str(self.x) }
        xmlattrs.update(attributes)

        xmlobject = etree.Element('node', xmlattrs)

        for (key, value) in self.tags.items():
            tag = etree.Element('tag', { 'k':key, 'v':value })
            xmlobject.append(tag)
        
        return etree.tostring(xmlobject, encoding='unicode')



class OsmWay(OsmGeometry):
    def __init__(self):
        super().__init__()
        self.points = []


    def replacejwithi(self, i, j):
        self.points = [ i if x == j else x for x in self.points ]
        j.removeparent(self)
        i.addparent(self)


    def to_xml(self, attributes):
        xmlattrs = { 'visible':'true', 'id':str(self.id) }
        xmlattrs.update(attributes)

        xmlobject = etree.Element('way', xmlattrs)

        for node in self.points:
            nd = etree.Element('nd', { 'ref':str(node.id) })
            xmlobject.append(nd)
        for (key, value) in self.tags.items():
            tag = etree.Element('tag', { 'k':key, 'v':value })
            xmlobject.append(tag)

        return etree.tostring(xmlobject, encoding='unicode')



class OsmRelation(OsmGeometry):
    def __init__(self):
        super().__init__()
        self.members = []


    def replacejwithi(self, i, j):
        self.members = [ (i, x[1]) if x[0] == j else x for x in self.members ]
        j.removeparent(self)
        i.addparent(self)


    def to_xml(self, attributes):
        xmlattrs = { 'visible':'true', 'id':str(self.id) }
        xmlattrs.update(attributes)

        xmlobject = etree.Element('relation', xmlattrs)

        for (member, role) in self.members:
            member = etree.Element('member', { 'type':'way', 'ref':str(member.id), 'role':role })
            xmlobject.append(member)

        tag = etree.Element('tag', { 'k':'type', 'v':'multipolygon' })
        xmlobject.append(tag)
        for (key, value) in self.tags.items():
            tag = etree.Element('tag', { 'k':key, 'v':value })
            xmlobject.append(tag)

        return etree.tostring(xmlobject, encoding='unicode')




# -*- coding: utf-8 -*-

from .translation_base_class import TranslationBase
from .datawriter_base_class import DataWriterBase
from .osm_datawriter import OsmDataWriter
from .pbf_datawriter import PbfDataWriter
from .ogr_datasource import OgrDatasource
from .osm_data import OsmData




# -*- coding: utf-8 -*-

''' ogr2pbf

This program takes any vector data understandable by OGR and outputs an OSM or
PBF file with that data.

By default tags will be naively copied from the input data. Hooks are provided
so that, with a little python programming, you can translate the tags however
you like. More hooks are provided so you can filter or even modify the features
themselves.

To use the hooks, create a file called myfile.py and run ogr2pbf.py -t myfile.
This file should define a class derived from TranslationBase where the hooks
you want to use are overridden.

The program will use projection metadata from the source, if it has any. If
there is no projection information, or if you want to override it, you can use
-e or -p to specify an EPSG code or Proj.4 string, respectively. If there is no
projection metadata and you do not specify one, EPSG:4326 will be used (WGS84
latitude-longitude)

For additional usage information, run ogr2pbf --help

Copyright (c) 2020 Roel Derickx <roel.derickx AT gmail>

Based on ogr2osm:
Copyright (c) 2012-2013 Paul Norman <penorman@mac.com>, Sebastiaan Couwenberg
<sebastic@xs4all.nl>, The University of Vermont <andrew.guertin@uvm.edu>

Released under the MIT license: http://opensource.org/licenses/mit-license.php

ogr2osm is based very heavily on code released under the following terms:
(c) Iván Sánchez Ortega, 2009
<ivan@sanchezortega.es>
###############################################################################
#  "THE BEER-WARE LICENSE":                                                   #
#  <ivan@sanchezortega.es> wrote this file. As long as you retain this notice #
#  you can do whatever you want with this stuff. If we meet some day, and you #
#  think this stuff is worth it, you can buy me a beer in return.             #
###############################################################################
'''

import sys, os, argparse, logging, inspect

from .translation_base_class import TranslationBase
from .osm_geometries import OsmId
from .ogr_datasource import OgrDatasource
from .osm_data import OsmData
from .osm_datawriter import OsmDataWriter
from .pbf_datawriter import PbfDataWriter

logging.basicConfig(format="%(message)s", level = logging.DEBUG)

def parse_commandline():
    parser = argparse.ArgumentParser()
    
    #parser.add_argument("-v", "--verbose", dest="verbose", action="store_true")
    #parser.add_argument("-d", "--debug-tags", dest="debugTags", action="store_true",
    #                    help="Output the tags for every feature parsed.")
    parser.add_argument("-t", "--translation", dest="translationmodule",
                        metavar="TRANSLATION",
                        help="Select the attribute-tags translation method. See " +
                             "the translations/ directory for valid values.")
    # datasource options
    parser.add_argument("--encoding", dest="encoding",
                        help="Encoding of the source file. If specified, overrides " +
                             "the default of %(default)s", default="utf-8")
    parser.add_argument("--sql", dest="sqlQuery", type=str, default=None,
                        help="SQL query to execute on a PostgreSQL source")
    parser.add_argument("--no-memory-copy", dest="noMemoryCopy", action="store_true",
                        help="Do not make an in-memory working copy")
    # input projection parameters
    parser.add_argument("-e", "--epsg", dest="sourceEPSG", type=int, metavar="EPSG_CODE",
                        help="EPSG code of source file. Do not include the " +
                             "'EPSG:' prefix. If specified, overrides projection " +
                             "from source metadata if it exists.")
    parser.add_argument("-p", "--proj4", dest="sourcePROJ4", type=str, metavar="PROJ4_STRING",
                        help="PROJ.4 string. If specified, overrides projection " +
                             "from source metadata if it exists.")
    parser.add_argument("--gis-order", dest="gisorder", action="store_true",
                        help="Consider the source coordinates to be in traditional GIS order")
    # precision options
    parser.add_argument("--significant-digits", dest="significantDigits", type=int,
                        help="Number of decimal places for coordinates to output " +
                             "(default: %(default)s)",
                        default=9)
    parser.add_argument("--rounding-digits", dest="roundingDigits", type=int,
                        help="Number of decimal places for rounding when snapping " +
                             "nodes together (default: %(default)s)",
                        default=7)
    # transformation options
    parser.add_argument("--split-ways", dest="maxNodesPerWay", type=int, default=1800,
                        help="Split ways with more than the specified number of nodes. " +
                             "Defaults to %(default)s. Any value below 2 - do not split.")
    # ID generation options
    parser.add_argument("--id", dest="id", type=int, default=0,
                        help="ID to start counting from for the output file. " +
                             "Defaults to %(default)s.")
    parser.add_argument("--idfile", dest="idfile", type=str, default=None,
                        help="Read ID to start counting from from a file.")
    parser.add_argument("--saveid", dest="saveid", type=str, default=None,
                        help="Save last ID after execution to a file.")
    parser.add_argument("--positive-id", dest="positiveId", action="store_true",
                        help=argparse.SUPPRESS) # can cause problems when used inappropriately
    # output file options
    parser.add_argument("-o", "--output", dest="outputFile", metavar="OUTPUT",
                        help="Set destination .osm file name and location.")
    parser.add_argument("-f", "--force", dest="forceOverwrite", action="store_true",
                        help="Force overwrite of output file.")
    parser.add_argument("--osm", dest="osm", action="store_true",
                        help="Write the output as an OSM file in stead of a PBF file")
    parser.add_argument("--no-upload-false", dest="noUploadFalse", action="store_true",
                        help="Omit upload=false from the completed file to suppress " +
                             "JOSM warnings when uploading.")
    parser.add_argument("--never-download", dest="neverDownload", action="store_true",
                        help="Prevent JOSM from downloading more data to this file.")
    parser.add_argument("--never-upload", dest="neverUpload", action="store_true",
                        help="Completely disables all upload commands for this file " +
                             "in JOSM, rather than merely showing a warning before " +
                             "uploading.")
    parser.add_argument("--locked", dest="locked", action="store_true",
                        help="Prevent any changes to this file in JOSM, " +
                             "such as editing or downloading, and also prevents uploads. " +
                             "Implies upload=\"never\" and download=\"never\".")
    parser.add_argument("--add-version", dest="addVersion", action="store_true",
                        help=argparse.SUPPRESS) # can cause problems when used inappropriately
    parser.add_argument("--add-timestamp", dest="addTimestamp", action="store_true",
                        help=argparse.SUPPRESS) # can cause problems when used inappropriately
    # required source file
    parser.add_argument("source", metavar="DATASOURCE",
                        help="DATASOURCE can be a file path or a org PostgreSQL connection " +
                             "string such as: \"PG:dbname=pdx_bldgs user=emma host=localhost\" " +
                             "(including the quotes)")
    params = parser.parse_args()
    
    if params.outputFile:
        params.outputFile = os.path.realpath(params.outputFile)

    # check consistency of parameters
    if params.source.startswith('PG:'):
        if not params.outputFile:
            parser.error("ERROR: An output file must be explicitly specified " +
                         "when using a database source")
        if not params.sqlQuery:
            parser.error("ERROR: You must specify a query with --sql when using a database source")
    else:
        if not params.outputFile:
            (base, ext) = os.path.splitext(os.path.basename(params.source))
            output_ext = ".osm.pbf"
            if params.osm:
                output_ext = ".osm"
            params.outputFile = os.path.join(os.getcwd(), base + output_ext)
        else:
            (base, ext) = os.path.splitext(os.path.basename(params.outputFile))
            if params.osm and ext.lower() == '.pbf':
                logging.warning("WARNING: You specified OSM output with --osm " +
                                "but the outputfile has extension .pbf, " +
                                "ignoring --osm parameter")
                params.osm = False
            elif not params.osm and ext.lower() == '.osm':
                logging.warning("WARNING: You didn't specify OSM output with --osm " +
                                "but the outputfile has extension .osm, " +
                                "automatically setting --osm parameter")
                params.osm = True
        if params.sqlQuery:
            logging.warning("WARNING: You specified a query with --sql " +
                            "but you are not using a database source")

    if not params.forceOverwrite and os.path.exists(params.outputFile):
        parser.error("ERROR: output file '%s' exists" % (params.outputFile))
    
    return params


def load_translation_object(translation_module):
    import_translation_module = None
    translation_object = None

    if translation_module:
        # add dirs to path if necessary
        (root, ext) = os.path.splitext(translation_module)
        if os.path.exists(translation_module) and ext == '.py':
            # user supplied translation file directly
            sys.path.insert(0, os.path.dirname(root))
        else:
            # first check translations in the subdir translations of cwd
            sys.path.insert(0, os.path.join(os.getcwd(), "translations"))
            # then check subdir of script dir
            sys.path.insert(1, os.path.join(os.path.dirname(__file__), "translations"))
            # (the cwd will also be checked implicityly)

        # strip .py if present, as import wants just the module name
        if ext == '.py':
            translation_module = os.path.basename(root)

        imported_module = None
        try:
            imported_module = __import__(translation_module)
        except ImportError as e:
            logging.error("Could not load translation method '%s'. Translation "
                          "script must be in your current directory, or in the "
                          "translations/ subdirectory of your current or ogr2osm.py "
                          "directory. The following directories have been considered: %s"
                          % (translation_module, str(sys.path)))
        except SyntaxError as e:
            logging.error("Syntax error in '%s'. Translation script is malformed:\n%s"
                          % (translation_module, e))
    
        for class_name in [ d for d in dir(imported_module) \
                                    if d != 'TranslationBase' and not d.startswith('__') ]:
            translation_class = getattr(imported_module, class_name)

            if inspect.isclass(translation_class) and \
               issubclass(translation_class, TranslationBase):
                logging.info('Found valid translation class %s' % class_name)
                setattr(sys.modules[__name__], class_name, translation_class)
                translation_object = translation_class()
                break

    if not translation_object:
        logging.info('Using default translations')
        translation_object = TranslationBase()
    
    return translation_object


def main():
    params = parse_commandline()

    translation_object = load_translation_object(params.translationmodule)

    OsmId.set_id(params.id, params.positiveId)
    if params.idfile:
        OsmId.load_id(params.idfile)

    logging.info("Preparing to convert '%s' to '%s'." % (params.source, params.outputFile))

    osmdata = OsmData(translation_object, params.significantDigits, params.roundingDigits, \
                      params.maxNodesPerWay)
    # create datasource and process data
    datasource = OgrDatasource(translation_object, \
                               params.sourcePROJ4, params.sourceEPSG, params.gisorder)
    datasource.open_datasource(params.source, not params.noMemoryCopy)
    datasource.set_query(params.sqlQuery)
    osmdata.process(datasource)
    #create datawriter and write OSM data
    datawriter = None
    if params.osm:
        datawriter = OsmDataWriter(params.outputFile, params.neverUpload, params.noUploadFalse, \
                                   params.neverDownload, params.locked, params.addVersion, \
                                   params.addTimestamp)
    else:
        datawriter = PbfDataWriter(params.outputFile, params.addVersion, params.addTimestamp)
    osmdata.output(datawriter)

    if params.saveid:
        OsmId.save_id(params.saveid)




# -*- coding: utf-8 -*-

import logging
from osgeo import ogr
from osgeo import osr

from .osm_geometries import OsmPoint, OsmWay, OsmRelation

class OsmData:
    def __init__(self, translation, significant_digits=9, rounding_digits=7, max_points_in_way=1800):
        # options
        self.translation = translation
        self.significant_digits = significant_digits
        self.rounding_digits = rounding_digits
        self.max_points_in_way = max_points_in_way
        
        self.__nodes = []
        self.__ways = []
        self.__relations = []
        self.__linestring_points = {}
        self.__long_ways_from_polygons = set()


    # This function builds up a dictionary with the source data attributes
    # and passes them to the filter_tags function, returning the result.
    def __get_feature_tags(self, ogrfeature):
        feature_def = ogrfeature.GetDefnRef()
        field_names = []
        for i in range(feature_def.GetFieldCount()):
            field_names.append(feature_def.GetFieldDefn(i).GetNameRef())

        tags = {}
        for j in range(len(field_names)):
            # The field needs to be put into the appropriate encoding and
            # leading or trailing spaces stripped
            tags[field_names[j]] = ogrfeature.GetFieldAsString(j).strip()
        return self.translation.filter_tags(tags)


    def __trunc_significant(self, n):
        return int(round(n * 10**self.significant_digits)) * 10**-self.significant_digits
    
    
    def __round_number(self, n):
        return int(round(n * 10**self.rounding_digits))
    
    
    def __parse_point(self, ogrgeometry):
        x = self.__trunc_significant(ogrgeometry.GetX())
        y = self.__trunc_significant(ogrgeometry.GetY())
        node = OsmPoint(x, y)
        self.__nodes.append(node)
        return node


    def __parse_linestring(self, ogrgeometry):
        way = OsmWay()
        # LineString.GetPoint() returns a tuple, so we can't call parsePoint on it
        # and instead have to create the point ourself
        for i in range(ogrgeometry.GetPointCount()):
            (x, y, z_unused) = ogrgeometry.GetPoint(i)
            rx = self.__round_number(x)
            ry = self.__round_number(y)
            if (rx, ry) in self.__linestring_points:
                mypoint = self.__linestring_points[(rx, ry)]
            else:
                mypoint = OsmPoint(self.__trunc_significant(x), self.__trunc_significant(y))
                self.__nodes.append(mypoint)
                self.__linestring_points[(rx, ry)] = mypoint
            way.points.append(mypoint)
            mypoint.addparent(way)
        self.__ways.append(way)
        return way


    def __parse_polygon(self, ogrgeometry):
        # Special case polygons with only one ring. This does not (or at least
        # should not) change behavior when simplify relations is turned on.
        if ogrgeometry.GetGeometryCount() == 0:
            logging.warning("Polygon with no rings?")
        elif ogrgeometry.GetGeometryCount() == 1:
            result = self.__parse_linestring(ogrgeometry.GetGeometryRef(0))
            if len(result.points) > self.max_points_in_way:
                self.__long_ways_from_polygons.add(result)
            return result
        else:
            relation = OsmRelation()
            try:
                exterior = self.__parse_linestring(ogrgeometry.GetGeometryRef(0))
                exterior.addparent(relation)
            except:
                logging.warning("Polygon with no exterior ring?")
                return None
            relation.members.append((exterior, "outer"))
            for i in range(1, ogrgeometry.GetGeometryCount()):
                interior = self.__parse_linestring(ogrgeometry.GetGeometryRef(i))
                interior.addparent(relation)
                relation.members.append((interior, "inner"))
            self.__relations.append(relation)
            return relation


    def __parse_collection(self, ogrgeometry):
        # OGR MultiPolygon maps easily to osm multipolygon, so special case it
        # TODO: Does anything else need special casing?
        geometry_type = ogrgeometry.GetGeometryType()
        if geometry_type in [ ogr.wkbMultiPolygon, ogr.wkbMultiPolygon25D ]:
            if ogrgeometry.GetGeometryCount() > 1:
                relation = OsmRelation()
                for polygon in range(ogrgeometry.GetGeometryCount()):
                    ext_geom = ogrgeometry.GetGeometryRef(polygon).GetGeometryRef(0)
                    exterior = self.__parse_linestring(ext_geom)
                    exterior.addparent(relation)
                    relation.members.append((exterior, "outer"))
                    for i in range(1, ogrgeometry.GetGeometryRef(polygon).GetGeometryCount()):
                        int_geom = ogrgeometry.GetGeometryRef(polygon).GetGeometryRef(i)
                        interior = self.__parse_linestring(int_geom)
                        interior.addparent(relation)
                        relation.members.append((interior, "inner"))
                self.__relations.append(relation)
                return [ relation ]
            else:
               return [ self.__parse_polygon(ogrgeometry.GetGeometryRef(0)) ]
        elif geometry_type in [ ogr.wkbMultiLineString, ogr.wkbMultiLineString25D ]:
            geometries = []
            for linestring in range(ogrgeometry.GetGeometryCount()):
                geometries.append(self.__parse_linestring(ogrgeometry.GetGeometryRef(linestring)))
            return geometries
        else:
            relation = OsmRelation()
            for i in range(ogrgeometry.GetGeometryCount()):
                member = self.__parse_geometry(ogrgeometry.GetGeometryRef(i))
                member.addparent(relation)
                relation.members.append((member, "member"))
            self.__relations.append(relation)
            return [ relation ]
    
    
    def __parse_geometry(self, ogrgeometry):
        osmgeometries = []
        
        geometry_type = ogrgeometry.GetGeometryType()

        if geometry_type in [ ogr.wkbPoint, ogr.wkbPoint25D ]:
            osmgeometries.append(self.__parse_point(ogrgeometry))
        elif geometry_type in [ ogr.wkbLineString, ogr.wkbLinearRing, ogr.wkbLineString25D ]:
            # ogr.wkbLinearRing25D does not exist
            osmgeometries.append(self.__parse_linestring(ogrgeometry))
        elif geometry_type in [ ogr.wkbPolygon, ogr.wkbPolygon25D ]:
            osmgeometries.append(self.__parse_polygon(ogrgeometry))
        elif geometry_type in [ ogr.wkbMultiPoint, ogr.wkbMultiLineString, ogr.wkbMultiPolygon, \
                               ogr.wkbGeometryCollection, ogr.wkbMultiPoint25D, \
                               ogr.wkbMultiLineString25D, ogr.wkbMultiPolygon25D, \
                               ogr.wkbGeometryCollection25D ]:
            osmgeometries.extend(self.__parse_collection(ogrgeometry))
        else:
            logging.warning("Unhandled geometry, type %s" % str(geometry_type))

        return osmgeometries


    def add_feature(self, ogrfeature, reproject = lambda geometry: None):
        if ogrfeature is None:
            return
        
        #reproject = self.__get_reprojection_func(spatial_ref)
        
        ogrfilteredfeature = self.translation.filter_feature(ogrfeature, reproject)
        if ogrfilteredfeature is None:
            return
        
        ogrgeometry = ogrfilteredfeature.GetGeometryRef()
        if ogrgeometry is None:
            return
        
        feature_tags = self.__get_feature_tags(ogrfilteredfeature)
        reproject(ogrgeometry)
        osmgeometries = self.__parse_geometry(ogrgeometry)

        for osmgeometry in [ geom for geom in osmgeometries if geom ]:
            osmgeometry.add_tags(feature_tags)
            
            self.translation.process_feature_post(osmgeometry, ogrfilteredfeature, ogrgeometry)


    def merge_points(self):
        logging.debug("Merging points")

        # Make list of Points at each location
        logging.debug("Making list")
        pointcoords = {}
        for i in self.__nodes:
            rx = self.__round_number(i.x)
            ry = self.__round_number(i.y)
            if (rx, ry) in pointcoords:
                pointcoords[(rx, ry)].append(i)
            else:
                pointcoords[(rx, ry)] = [i]

        # Use list to get rid of extras
        logging.debug("Checking list")
        for (location, pointsatloc) in pointcoords.items():
            if len(pointsatloc) > 1:
                for point in pointsatloc[1:]:
                    for parent in set(point.get_parents()):
                        parent.replacejwithi(pointsatloc[0], point)
                    if len(point.get_parents()) == 0:
                        self.__nodes.remove(point)


    def merge_way_points(self):
        logging.debug("Merging duplicate points in ways")

        # Remove duplicate points from ways,
        # a duplicate has the same id as its predecessor
        for way in self.__ways:
            previous_id = None
            merged_points = []

            for node in way.points:
                if previous_id == None or previous_id != node.id:
                    merged_points.append(node)
                    previous_id = node.id

            if len(merged_points) > 0:
                way.points = merged_points
            else:
                # TODO delete way? also delete from parent relation?
                pass


    def __split_way(self, way, is_way_in_relation):
        new_points = [ way.points[i:i + self.max_points_in_way] \
                               for i in range(0, len(way.points), self.max_points_in_way - 1) ]
        new_ways = [ way ] + [ OsmWay() for i in range(len(new_points) - 1) ]

        if not is_way_in_relation:
            way_tags = way.tags

            for new_way in new_ways:
                if new_way != way:
                    new_way.add_tags(way_tags)
                    self.__ways.append(new_way)

        for new_way, points in zip(new_ways, new_points):
            new_way.points = points
            if new_way.id != way.id:
                for point in points:
                    point.removeparent(way)
                    point.addparent(new_way)

        return new_ways


    def __merge_into_new_relation(self, way_parts):
        new_relation = OsmRelation()
        self.__relations.append(new_relation)
        new_relation.members = [ (way, "outer") for way in way_parts ]
        for way in way_parts:
            way.addparent(new_relation)


    def __split_way_in_relation(self, rel, way_parts):
        way_roles = [ m[1] for m in rel.members if m[0] == way_parts[0] ]
        way_role = "" if len(way_roles) == 0 else way_roles[0]
        for way in way_parts[1:]:
            rel.members.append((way, way_role))


    def split_long_ways(self):
        if self.max_points_in_way < 2:
            # pointless :-)
            return
        
        logging.debug("Splitting long ways")

        for way in self.__ways:
            is_way_in_relation = len([ p for p in way.get_parents() if type(p) == OsmRelation ]) > 0
            if len(way.points) > self.max_points_in_way:
                way_parts = self.__split_way(way, is_way_in_relation)
                if not is_way_in_relation:
                    if way in self.__long_ways_from_polygons:
                        self.__merge_into_new_relation(way_parts)
                else:
                    for rel in way.get_parents():
                        self.__split_way_in_relation(rel, way_parts)


    def process(self, datasource):
        for i in range(datasource.get_layer_count()):
            (layer, reproject) = datasource.get_layer(i)
            
            if layer:
                for i in range(layer.GetFeatureCount()):
                    ogrfeature = layer.GetNextFeature()
                    
                    if ogrfeature:
                        self.add_feature(ogrfeature, reproject)

        self.merge_points()
        self.merge_way_points()
        self.split_long_ways()


    class DataWriterContextManager:
        def __init__(self, datawriter):
            self.datawriter = datawriter
        
        def __enter__(self):
            self.datawriter.open()
            return self.datawriter
        
        def __exit__(self, exception_type, value, traceback):
            self.datawriter.close()


    def output(self, datawriter):
        self.translation.process_output(self.__nodes, self.__ways, self.__relations)
        
        with self.DataWriterContextManager(datawriter) as dw:
            dw.write_header()
            dw.write_nodes(self.__nodes)
            dw.write_ways(self.__ways)
            dw.write_relations(self.__relations)
            dw.write_footer()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

exec(open("ogr2pbf/version.py").read())

import setuptools

with open('README.md', 'r', encoding='utf-8') as fh:
    README = fh.read()

setuptools.setup(
    name="ogr2pbf",
    version=__version__,
    license=__license__,
    author=__author__,
    author_email="ogr2pbf.pypi@derickx.be",
    description="A tool for converting ogr-readable files like shapefiles into .osm or .pbf data",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/roelderickx/ogr2pbf",
    packages=setuptools.find_packages(),
    python_requires='>=3.7',
    entry_points={
        'console_scripts': ['ogr2pbf = ogr2pbf.ogr2pbf:main']
    },
    classifiers=[
        'Environment :: Console',
        'Topic :: Scientific/Engineering :: GIS',
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)

